/**
 * Represents a node in a binary tree.
 */
class TreeNode<E> {
  private E data;
  public TreeNode<E> left;
  public TreeNode<E> right;

  /**
   * Creates a node and saves the input data.
   * 
   * @param data to save at the node.
   */
  public TreeNode(E data) {
    this.data = data;
    left = null;
    right = null;
  }

  /**
   * Retrieves the data saved at the node.
   * 
   * @return
   */
  public E get() {
    return data;
  }

  /**
   * @return a String representation of this node.
   */
  public String toString() {
    return data.toString();
  }
}

/**
 * Represents a binary tree, where each node holds
 * some data of type E.
 */
public class BinaryTree<E> {
  public TreeNode<E> root;

  /**
   * Constructs an empty binary tree.
   */
  public BinaryTree() {
    root = null;
  }

  /**
   * Returns a string representation of the subtree.
   * 
   * @param padding - spacing and vertical bars to print before the node.
   * @param node    - the root of the subtree to print.
   * @return appended line to print for this node (and child nodes).
   */
  public String toStringHelper(String padding, TreeNode<E> node) {
    if (node == null) {
      return "";
    }

    String result = padding + "└──(" + node.toString() + ")" + "\n";

    padding += "│    ";
    result += toStringHelper(padding, node.left);
    result += toStringHelper(padding, node.right);

    return result;
  }

  /**
   * @return the height of the tree
   */
  public int height() {
    return height(root);
  }

  private int height(TreeNode<E> node) {
    if (node == null) return -1;
    int leftHeight = height(node.left);
    int rightHeight = height(node.right);
    return 1 + Math.max(leftHeight, rightHeight);
  }

  /**
   * @return a pre-order traversal of nodes of this tree
   */
  public String visitPreOrder() {
    return visitPreOrder(root);
  }

  private String visitPreOrder(TreeNode<E> node) {
    String result = "";
    if (node == null) return result;
    result += node.get() + " ";
    result += visitPreOrder(node.left);
    result += visitPreOrder(node.right);
    return result;
  }

  /**
   * @return an in-order traversal of nodes of this tree
   */
  public String visitInOrder() {
    return visitInOrder(root);
  }

  private String visitInOrder(TreeNode<E> node) {
    String result = "";
    if (node == null) return result;
    result += visitInOrder(node.left);
    result += node.get() + " ";
    result += visitInOrder(node.right);
    return result;
  }

  /**
   * @return a post-order traversal of nodes of this tree
   */
  public String visitPostOrder() {
    return visitPostOrder(root);
  }

  private String visitPostOrder(TreeNode<E> node) {
    String result = "";
    if (node == null) return result;
    result += visitPostOrder(node.left);
    result += visitPostOrder(node.right);
    result += node.get() + " ";
    return result;
  }

  public static void main(String[] args) {
    BinaryTree<String> tree = new BinaryTree<>();
    tree.root = new TreeNode<String>("R");

    TreeNode<String> f = new TreeNode<>("F");
    TreeNode<String> c = new TreeNode<>("C");
    tree.root.left = f;
    tree.root.right = c;
    TreeNode<String> j = new TreeNode<>("J");
    j.left = new TreeNode<>("M"); 
    j.right = new TreeNode<>("P");
    TreeNode<String> a = new TreeNode<>("A");
    a.left = new TreeNode<>("T");
    a.right = new TreeNode<>("X");
    f.left = j;
    f.right = a;
    TreeNode<String> g = new TreeNode<>("G");
    g.left = new TreeNode<>("B");
    g.right = new TreeNode<>("E");
    c.left = new TreeNode<>("D");
    c.right = g;

    System.out.println("Tree height: " + tree.height());
    System.out.println("Pre-order traversal: " + tree.visitPreOrder());
    System.out.println("In-order traversal: " + tree.visitInOrder());
    System.out.println("Post-order traversal: " + tree.visitPostOrder());
    System.out.println(tree.toStringHelper("", tree.root));
  }
}